theory Examples
imports "../TermFun" Loga
begin

section {* Exmaple - An Odd Even/Odd Function *}

text {* The first argument is a flag where @{term "True"} indicates that we
want to check for even numbers and @{term "False"} that we want to check for
odd numbers. *}
function evenodd :: "bool \<Rightarrow> nat \<Rightarrow> bool"
where
  "evenodd True 0 = False"
| "evenodd False x = (\<not> evenodd True x)"
| "evenodd True (Suc x) = evenodd False x"
  by (pat_completeness) auto

text {* As in @{theory Loga}, we first generate a TRS representation of the
involved functions. For logical negation @{const "Not"}, we use two equations
from the standard library. *}
define_trs not_True_eq_False not_False_eq_True (* order is important, why? *)
define_trs (no_sim) evenodd.psimps

text {* The resulting TRS has the name @{term "evenodd.trs"} and its definition
is provided via the lemma @{thm[source] evenodd.trs_def}. *}
term evenodd.trs
thm evenodd.trs_def

text {* The termination proof involves the same steps as for @{theory Loga}. *}
termination evenodd
  txt {* First we reduce termination of the function to termination of the TRS.  *}
  apply (trs_reduce evenodd)
  txt {* Then we obtain a representation of the TRS as finite list of rules. *}
  apply expand_trs_problem
  txt {* We export the resulting list of rules into an XML file. *}
  export_termination_problem "evenodd.xml"
  txt {* A corresponding XML certificate, produced by some external termination tool,
  is reconstructed *}
  apply (ceta_import_proof "proofs/evenodd.proof.xml")
  txt {* and checked. *}
  apply ceta_eval
done


section {* Example - Normalization of Conditionals (cf. Paulson 1986) *}

datatype exp = AT nat | IF exp exp exp

text {* Since a user-defined datatype is involved, we have to register it in order
to obtain an embedding function @{const "embed"} encoding values of the datatype by
first-order terms. *}
setup {* Term_Fun.register_datatypes [@{type_name exp}] *}
thm embed_exp.simps

function norm where
  "norm (AT a) = AT a"
| "norm (IF (AT a) y z) = IF (AT a) (norm y) (norm z)"
| "norm (IF (IF u v w) y z) = norm (IF u (IF v y z) (IF w y z))"
  by pat_completeness auto

text {* A TRS representation of @{const norm} is established using the
@{text define_trs} command. *}
define_trs (no_sim) norm.psimps

termination norm 
  apply (trs_reduce norm)
  apply expand_trs_problem
  export_termination_problem "norm.xml"
  apply (ceta_import_proof "proofs/norm.proof.xml")
  apply ceta_eval
done


section {* Example - Preprocessing of Formulas (Chaieb) *}

datatype num = C int | Bound nat | CN nat int num | Neg num | Add num num| Sub num num 
  | Mul int num | Floor num| CF int num num

text {* Obtain a term representation of the type @{typ "num"} *}
setup {* Term_Fun.register_datatypes [@{type_name num}] *}

text {* since int is not embedded, we will see dummy terms in embed_num.simps *}
thm embed_num.simps

datatype fm  = 
  T| F| Lt num| Le num| Gt num| Ge num| Eq num| NEq num| Dvd int num| NDvd int num|
  NOT fm| And fm fm|  Or fm fm| Imp fm fm| Iff fm fm| E fm| A fm

text {* as well as the type @{typ "fm"}. *}
setup {* Term_Fun.register_datatypes [@{type_name fm}] *}
thm embed_fm.simps

function (sequential) prep :: "fm \<Rightarrow> fm" where
  "prep (E T) = T"
| "prep (E F) = F"
| "prep (E (Or p q)) = Or (prep (E p)) (prep (E q))"
| "prep (E (Imp p q)) = Or (prep (E (NOT p))) (prep (E q))"
| "prep (E (Iff p q)) = Or (prep (E (And p q))) (prep (E (And (NOT p) (NOT q))))" 
| "prep (E (NOT (And p q))) = Or (prep (E (NOT p))) (prep (E(NOT q)))"
| "prep (E (NOT (Imp p q))) = prep (E (And p (NOT q)))"
| "prep (E (NOT (Iff p q))) = Or (prep (E (And p (NOT q)))) (prep (E(And (NOT p) q)))"
| "prep (E p) = E (prep p)"
| "prep (A (And p q)) = And (prep (A p)) (prep (A q))"
| "prep (A p) = prep (NOT (E (NOT p)))"
| "prep (NOT (NOT p)) = prep p"
| "prep (NOT (And p q)) = Or (prep (NOT p)) (prep (NOT q))"
| "prep (NOT (A p)) = prep (E (NOT p))"
| "prep (NOT (Or p q)) = And (prep (NOT p)) (prep (NOT q))"
| "prep (NOT (Imp p q)) = And (prep p) (prep (NOT q))"
| "prep (NOT (Iff p q)) = Or (prep (And p (NOT q))) (prep (And (NOT p) q))"
| "prep (NOT p) = NOT (prep p)"
| "prep (Or p q) = Or (prep p) (prep q)"
| "prep (And p q) = And (prep p) (prep q)"
| "prep (Imp p q) = prep (Or (NOT p) q)"
| "prep (Iff p q) = Or (prep (And p q)) (prep (And (NOT p) (NOT q)))"
| "prep p = p"
  by pat_completeness simp_all

text {* The remaining steps are ``canonical.'' *}
define_trs (no_sim) prep.psimps

termination prep
  apply (trs_reduce prep)
  apply expand_trs_problem
  export_termination_problem "prep.xml"
  apply (ceta_import_proof "proofs/prep.proof.xml")
  apply ceta_eval
done


section {* Example - Merge Sort *}

fun merge :: "nat list \<Rightarrow> nat list \<Rightarrow> nat list" where
  "merge (x#xs) (y#ys) = (case x \<le> y of
    True => x # merge xs (y#ys)
  |  False => y # merge (x#xs) ys)"
| "merge xs [] = xs"
| "merge [] ys = ys"
(*  by pat_completeness auto *)

function (sequential) msort :: "nat list \<Rightarrow> nat list" where
  "msort [] = []"
| "msort [x] = [x]"
| "msort xs = merge (msort (take (half (length xs)) xs))
	            (msort (drop (half (length xs)) xs))"
  by pat_completeness auto

(*
define_trs (no_sim) merge.psimps

text {* For @{const "merge"}, we are able to prove termination using our approach. *}
termination merge
  apply (trs_reduce merge)
  apply expand_trs_problem
  export_termination_problem "merge.xml"
  apply (ceta_import_proof "proofs/merge.proof.xml")
  apply ceta_eval
done
*)

define_trs merge.simps

define_trs (no_sim) msort.psimps
thm msort.trs_def

termination msort
  apply (trs_reduce msort)
  apply expand_trs_problem
  export_termination_problem "msort.xml"
  apply (ceta_import_proof "proofs/msort.proof.xml")
  apply ceta_eval
done

end